from class_runtime import Runtime
from class_profile import Profile
from class_course import Course
from class_assignment import Assignment

from easygui_mod import *
from sys import exit, platform
from os import path, makedirs


# ----------------------------------------------------------
# Settings
# ----------------------------------------------------------

PATH += 'coursegrade/'
FILENAME = '{}coursegrade_save.txt'.format(PATH)

CHOICE_NEW = '----Create new {}----'
NEW_CODE = '%new'

INVALID_CHARS = ['\\']

LETTER_TO_GRADE = {
    'A+': 90,
    'A': 85,
    'A-': 80,
    'B+': 77,
    'B': 73,
    'B-': 70,
    'C+': 67,
    'C': 63,
    'C-': 60,
    'D+': 57,
    'D': 53,
    'D-': 50,
    'F': 0}

MODE_TO_OBJECT = {
    'profile': Profile,
    'course': Course
    }


# ----------------------------------------------------------
# Miscellaneous functions
# ----------------------------------------------------------

def prettify_float(value):
    '''(float or int) -> str
    
    Remove all useless zeroes after a decimal and truncate to 2 decimals.
    
    >>> prettify_float(5.000)
    '5'
    >>> prettify_float(5.010)
    '5.01'
    '''
    
    if type(value) == int:
        return str(value)
    
    elif type(value) == float:
        value = str(value)
        decimal = value.find('.')
        
        first_half = value[:decimal]
        second_half = '{:.3}'.format(value[decimal:])
        second_half = second_half.rstrip('0').rstrip('.')
        
        return first_half + second_half
    
    
def convert_grade(s):
    
    if valid_number(s):
        return float(s) / 100
    else:
        s = s.split('/')
        return float(s[0].strip()) / float(s[1].strip())
    
# ----------------------------------------------------------    
# Validity functions
# ----------------------------------------------------------

def valid_string(s):
    '''(str) -> Boolean
    
    s is an input. Return True iff s is not empty and has no invalid chars.
    '''
    
    if s == '':
        return False
    
    for char in s:
        if char in INVALID_CHARS:
            return False
        
    return True


def valid_number(s):
    
    try:
        number = float(s)
        return True
    except:
        return False


def valid_fraction(s):
    
    grade = s.split('/')
    
    if len(grade) != 2:
        return False # Must be of the form 'digits/digits'
        
    try:
        numerator = float(grade[0].strip())
            
        try:
            denominator = float(grade[1].strip())
            
            if denominator <= 0:
                return False # Zero division error
            
            return True
        
        except(ValueError):
            return False # Failed to float the denominator
        
    except(ValueError):
        return False # Failed to float the numerator
    
    
def valid_letter(letter):
    
    grade = LETTER_TO_GRADE.get(letter)
    
    if grade != None:
        return True
    
    return False
 
 
def exists(existing, s):
    
    for item in existing.get_contents():
        if str(item) == s:
            return True
        
    return False

# ----------------------------------------------------------
# Core boxes
# ----------------------------------------------------------

def pick(context, title, mode):
    
    message = 'Pick the {} you want to work on.'.format(mode)
    name_to_item = {}
    choices = []
    
    # Grab the contents and duplicate them
    contents = context.get_contents()
    
    # Build the list of choices
    for item in contents:
        
        name = '{}  --  '.format(str(item))
        
        if mode == 'profile':
            
            CGPA = item.get_CGPA()
            
            if CGPA == -1:
                name += 'no CGPA yet'
            else:
                name += prettify_float(CGPA)
        
        elif mode == 'course':
            
            grade = item.get_grade()[0]
            
            if grade == -1:
                name += 'no grade yet'
            else:
                name += '{}%'.format(prettify_float(grade))
            
        elif mode == 'assignment':
            
            weight = item.get_weight()
            grade = item.get_grade()
            name += 'worth {}%'.format(prettify_float(weight))
            
            if grade == -1:
                name += ' , no grade yet'
            else:
                name += ' , received {}%'.format(prettify_float(100 * grade))
            
        name_to_item[name] = item
        choices.append(name)
    
    # Insert new choice
    new = CHOICE_NEW.format(mode)
    choices.append(new)
    
    choice = choicebox(message, title, choices, mode)
    
    if choice == None:
            return None
        
    elif choice == new:
        return NEW_CODE
    
    else:
        return name_to_item[choice]


def create_course_or_profile(context, title, mode):
    
    message = 'Enter the name of the {} to create.'.format(mode)
    
    name = ''
    
    while True:
        name = multenterbox(message, title, [''], [name])
        
        if name == None:
            return None
        
        name = name[0].strip()
        
        if not valid_string(name):
            msgbox('Please enter a valid name.', title)
                
        elif exists(context, name):
            msgbox('The {} "{}" already exists.'.format(mode, name), title)
        
        else:
            break
            
    created = MODE_TO_OBJECT[mode](name)
    return created
    

def create_assignment(course, title):
    
    free_weight = course.get_free_weight()
    
    if free_weight > 0:
        
        message = '\n'.join([
            'You must enter the name and weight now.',
            'You can enter the grade now, or leave it blank and enter it later.'])
        
        fields = ['Name: ', 'Weight, as %: ', 'Grade, as % or fraction: ']
        
        name, weight, grade = '', '', ''
        
        while True:
            
            values = [name, weight, grade]
            entered = multenterbox(message, title, fields, values)
            
            # User clicks Cancel.
            if entered == None:
                return None
            
            # Dismantle entered and decipher input
            name, weight, grade = entered
            name = name.strip()
            weight = weight.strip().replace('%', '')
            grade = grade.strip().replace('%', '')
            
            # Validity checks
            
            # Name is invalid
            if not valid_string(name):
                msgbox('Please enter a valid name.', title)
                
            # Name already exists
            elif exists(course, name):
                msgbox('The assignment "{}" already exists.'.format(name), title)            
            
            # Weight is invalid
            elif not valid_number(weight):
                msgbox('The weight must be a valid %.', title)
                
            # Weight is too low
            elif float(weight) <= 0:
                msgbox('The weight must be more than 0.', title)
                    
            # Weight is too high
            elif free_weight - float(weight) < 0:
                msgbox('That weight would bring the total over 100%.', title)            
            
            # Grade is invalid
            elif grade != '' and not valid_number(grade) and not valid_fraction(grade):
                msgbox('The grade must be a valid % or fraction.', title)
                
            # Grade is too low
            elif grade != '' and convert_grade(grade) < 0:
                msgbox('The grade must equal at least 0.', title)
                
            else:
                
                weight = float(weight)
                
                if grade == '':
                    grade = -1
                else:
                    grade = convert_grade(grade)
                    
                assignment = Assignment(name, weight, grade)
                return assignment
                
    else:
        msgbox("All of the course's weight is already accounted for.")
        return None
            

def del_box(victim, title):
    
    message = 'Are you sure you want to delete "{}"?'.format(str(victim))
    
    return boolbox(message, title)


def rename_box(victim, title):
    
    message = 'Enter the name you want to change it to.'
    name = str(victim)
    
    while True:
        name = multenterbox(message, title, [''], [name])
        
        if name == None:
            break
        
        name = name[0].strip()
        
        if not valid_string(name):
            msgbox('Please enter a valid name.', title)
            
        else:
            victim.rename(name)
            break


# ----------------------------------------------------------
# Other boxes
# ----------------------------------------------------------

def desired_box(course, title):
    unknown = 100 - course.get_grade()[2]
    
    # There is still some weight unaccounted for
    if unknown > 0:
        received = course.get_grade()[1]
        
        invalid_message = 'Please enter a valid % or letter grade.'

        display = ''
        
        message = 'Enter the % or letter grade you want and click OK to update.'
        
        while True:
            
            desired = multenterbox(message, title,
                                   ['I hope to get: '], [display])
            
            # User clicks Cancel
            if desired == None:
                break
            
            else:
                desired = desired[0].strip().replace('%', '')
                
                # Input is invalid
                if not valid_number(desired) and not valid_letter(desired):
                    msgbox(invalid_message, title)
                    
                else:
                        
                    # Input is a float
                    if valid_number(desired):
                        desired = float(desired)
                        display = prettify_float(desired)
                        
                    # Input is a letter grade
                    else:
                        display = desired[:]
                        desired = LETTER_TO_GRADE[desired]
                    
                    # Update message box   
                    required = 100 * ((desired - received) / unknown)
                    
                    if required < 0: # Goal would require negative grades
                        message = ''.join([
                            "Enter the % or letter grade you want",
                            ' and click OK to update.',
                            '\n\nCongratulations! You can no longer achieve that low a grade...'
                            ])
                    
                    else:
                        message = ''.join([
                            "Enter the % or letter grade you want",
                            ' and click OK to update.',
                            '\n\nAverage needed on remaining work: {}%'
                            ]).format(prettify_float(required))
    
    # All the weight is accounted for.        
    else:
        msgbox('All your grades are already in.', title)
        
        
def bonus_box(course, title):
    
    invalid_message = 'Please enter a valid %.'
    message = 'Enter the total bonus as a %.'    
    bonus = prettify_float(course.get_bonus())
    
    while True:
        
        bonus = multenterbox(message, title, [''], [bonus])
        
        if bonus == None:
            break
        
        else:
            bonus = bonus[0].strip().replace('%', '')
            
            if not valid_number(bonus):
                msgbox(invalid_message, title)
                
            else:
                bonus = float(bonus)
                course.set_bonus(bonus)
                break


def assignment_grade_box(assignment, title):
       
    message = 'Enter the grade as a % or fraction.'
    
    grade = assignment.get_grade()
    
    if grade == -1:
        grade = ''
    else:
        grade = prettify_float(100 * grade)
    
    while True:
        
        grade = multenterbox(message, title,
                             [''], [grade])
        
        if grade == None:
            break
        
        grade = grade[0].strip().replace('%', '')
        
        if not valid_number(grade) and not valid_fraction(grade):
            msgbox('Please enter a valid % or fraction.', title)
            
        elif convert_grade(grade) < 0:
            msgbox('The grade must equal at least 0.', title)
            
        else:
                
            grade = convert_grade(grade)
            assignment.set_grade(grade)
            break


def weight_box(course, assignment, title):
    
    free_weight = course.get_free_weight() + assignment.get_weight()
    
    message = 'Enter the weight as a %.'
    weight = prettify_float(assignment.get_weight())
    
    while True:
        
        weight = multenterbox(message, title, [''], [weight])
        
        if weight == None:
            break
        
        weight = weight[0].strip().replace('%', '')
        
        if not valid_number(weight):
            msgbox('Please enter a valid %.', title)
            
        elif free_weight - float(weight) < 0:
            msgbox('That weight would bring the total over 100%.', title)
            
        elif float(weight) < 0:
            msgbox('The weight must be at least 0.', title)
            
        else:
            weight = float(weight)
            assignment.set_weight(weight)
            break
        
        
# ----------------------------------------------------------
# Menu messages
# ----------------------------------------------------------

def profile_menu_message(profile):
    
    CGPA = profile.get_CGPA()
    
    if CGPA == -1:
        return 'CGPA: none yet'
    else:
        return 'CGPA: {}'.format(prettify_float(CGPA))
        
        
def course_menu_message(course):
    grade, received, known, GPA, letter = course.get_grade()
    bonus = course.get_bonus()
    
    if grade == -1:
        message = ['Grade: none yet']
        
    else:
        message = [
            'Grade: {}%'.format(prettify_float(grade)),
            "Earned: {}% out of a possible {}%".format(prettify_float(received),
                                                       prettify_float(known)),
            'GPA: {}'.format(GPA),
            'Letter: {}'.format(letter)
            ]
    
    # Insert bonus line
    if bonus > 0:
        message.insert(2, 'Bonus: {}%'.format(prettify_float(bonus)))
    
    return '\n'.join(message)
    

def assignment_menu_message(assignment):
    
    weight = assignment.get_weight()
    grade = assignment.get_grade()
    
    message = 'Weight: {}%\n'.format(prettify_float(weight))
    
    if grade != -1: # There is a grade to display
        message += 'Grade: {}%'.format(prettify_float(100 * grade))
    else:
        message += 'Grade: none yet'
        
    return message


# ----------------------------------------------------------
# Menus
# ----------------------------------------------------------

def main_menu(runtime):
    
    title = 'Coursegrade'
    
    while True:
        profile = pick(runtime, title, 'profile')
        
        if profile == None:
            exit()
        
        else:
            
            if profile == NEW_CODE:
                profile = create_course_or_profile(runtime, title, 'profile')
                if profile != None:
                    runtime.add_profile(profile)
                    
            if profile != None:
                profile_menu(runtime, profile)

    
def profile_menu(runtime, profile):
    
    choices = [
        'Courses',
        'Rename',
        'Delete',
        'Back',
        'Save and quit'
        ]
        
    while True:
        
        title = 'Profile: {}'.format(str(profile))
        message = '{}\n{}'.format(title, profile_menu_message(profile))        
        
        choice = buttonbox(message, title, choices)
        
        if choice == 'Back':
            break
        
        elif choice == 'Delete':
            if del_box(profile, title):
                runtime.del_profile(profile)
                break
            
        elif choice == 'Rename':
            rename_box(profile, title)
            
        elif choice == 'Save and quit':
            exit()
            
        elif choice == 'Courses':
            course = pick(profile, title, 'course')
            
            if course != None:
                
                if course == NEW_CODE:
                    course = create_course_or_profile(profile, title, 'course')
                    if course != None:
                        profile.add_course(course)
                
                if course != None:
                    course_menu(profile, course)
    

def course_menu(profile, course):
    
    choices = [
        'Assignments',
        'Desired final grade',
        'Bonus mark',
        'Rename',
        'Delete',
        'Back',
        'Save and quit'
        ]
    
    while True:
        
        title = 'Course: {}'.format(str(course))
        message = '{}\n{}'.format(title, course_menu_message(course))        
        
        choice = buttonbox(message, title, choices)
    
        if choice == 'Back':
            break
        
        elif choice == 'Delete':
            if del_box(course, title):
                profile.del_course(course)
                break
            
        elif choice == 'Rename':
            rename_box(course, title)            
            
        elif choice == 'Save and quit':
            exit()
               
        elif choice == 'Desired final grade':
            desired_box(course, title)
        
        elif choice == 'Bonus mark':
            bonus_box(course, title)            
        
        elif choice == 'Assignments':
            assignment = pick(course, title, 'assignment')
            
            if assignment != None:
                
                if assignment == NEW_CODE:
                    assignment = create_assignment(course, title)
                    if assignment != None:
                        course.add_assn(assignment)
                
                if assignment != None:
                    assignment_menu(course, assignment)
                    

def assignment_menu(course, assignment):
    
    choices = [
        'Enter grade',
        'Change weight',
        'Rename',
        'Delete',
        'Back',
        'Save and quit'
        ]
    
    while True:
        
        title = 'Assignment: {}'.format(str(assignment))
        message = '{}\n{}'.format(title, assignment_menu_message(assignment))        
        
        choice = buttonbox(message, title, choices)
        
        if choice == 'Back':
            break
        
        elif choice == 'Save and quit':
            exit()
            
        elif choice == 'Delete':
            if del_box(assignment, title):
                course.del_assn(assignment)
                break
            
        elif choice == 'Rename':
            rename_box(assignment, title)                
                
        elif choice == 'Enter grade':
            assignment_grade_box(assignment, title)
            
        elif choice == 'Change weight':
            weight_box(course, assignment, title)


# ----------------------------------------------------------
# I/O functions
# ----------------------------------------------------------

def file_read(filename):
    '''() -> Runtime
    
    Read the save file at filename and return the runtime.
    '''
    
    runtime = Runtime()
    
    try:
        file = open(FILENAME, 'r')
        
        # Each block's first line is the profile name.
        while True:
            line = file.readline()
            
            if line == '':
                break
            
            else:
                line = line.strip()
                profile = Profile(line)
        
            # Each inner block's first line will be the course name.
            while True:
                line = file.readline().strip()
                
                if line == '':
                    break
                
                else:
                    line = line.strip()
                    name = ' '.join(line.split()[:-1])
                    bonus = float(line.split()[-1])
                    course = Course(name)
                    course.set_bonus(bonus)
                    
                    # After the course name comes a block of assignment lines.
                    while True:
                        line = file.readline().strip()
                        
                        if line == '':
                            break
                        
                        line = line.split()
                        name = ' '.join(line[:-2])
                        weight, grade = float(line[-2]), float(line[-1])
                        assn = Assignment(name, weight, grade)
                        
                        course.add_assn(assn)
                        
                    profile.add_course(course)
                
            runtime.add_profile(profile)
            
        file.close()
        return runtime
    
    except:
        return runtime

            
def file_save(filename, runtime):
    '''    
    Save runtime to a file.
    '''
    
    if not path.exists(PATH):
        makedirs(PATH)
    
    try:
        file = open(filename, 'w')
        
        students = runtime.get_contents()
        
        for student in students:
            courses = student.get_contents()
            courses_text = ''
        
            for course in courses:
                assignments = course.get_contents()
                assignment_text = ''
                
                for assignment in assignments:
                    assignment_text += '{} {} {}\n'.format(str(assignment),
                        str(assignment.get_weight()), str(assignment.get_grade()))
                
                courses_text += '{} {}\n{}\n'.format(str(course),
                                    str(course.get_bonus()), assignment_text)
                
            student_text = '{}\n{}\n'.format(str(student), courses_text)
            
            file.write(student_text)
                       
        file.close()
        msgbox('Your info was saved. :)', 'Coursegrade')
        
    except(FileNotFoundError):
        input('Unable to save! :(', 'Coursegrade')


# ----------------------------------------------------------
# Program itself
# ----------------------------------------------------------

runtime = file_read(FILENAME)

try:
    main_menu(runtime)
except(SystemExit):
    file_save(FILENAME, runtime)